/* 
   Copyright (c) 2004-2005 by Igor B. Kuznetsov. All Rights Reserved.
   IKuznetsov@albany.edu

   Licensor hereby grants to any person a permission to use, copy and distribute any part of this software
   program for educational and non-profit purposes, without fee, and without a written agreement,
   provided that this copyright notice appears in all copies. You cannot modify this source code without 
   an explicit written permission of the author. 

   Any unauthorized commercial distribution of any part of this software program is strictly prohibited. 

   This software is provided on an "AS IS" BASIS and WITHOUT WARRANTY, either express or implied, 
   including, without limitation, the warranties of NON-INFRINGEMENT, MERCHANTABILITY 
   or FITNESS FOR A PARTICULAR PURPOSE. THE ENTIRE RISK AS TO THE QUALITY OF THE SOFTWARE IS WITH YOU.

   Limitation of Liability. Under no circumstances and under no legal theory shall the Licensor be liable 
   to any person for any direct, indirect, special, incidental, or consequential damages of any character 
   arising as a result of the use of this software. 
*/

/* 
   NBIAS. A version for DNA sequences.
   This program uses discrete runs of user-specified nucleotides to identify window size by 
   merging the runs into one fragment. The significance is computed as P(X>=k|w) 
   using the scan statistic approximation from Eq.4.3 in Glaz et al.
*/

/* 
   Revision history:

   Dec-28-04: random sequences are generated by sampling from initial large pool of residues generated with given
   nucelotide frequencies. 

   Feb-09-05: sliding window of user-specified length added. Final window size is determined by merging 
   adjacent windows with p-value <= threshold.
   Prints FASTA file with masked sequence.
*/

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <time.h>
#include <math.h>
#include <limits.h>
#include <float.h>

#include <string>
#include <iostream>
#include <fstream>
#include <vector>
#include <algorithm>
#include <map>
#include <stdexcept>
//#include <new.h>

#define RAN_OFF 0     // Random sequences are off
#define RAN_FREQ 1    // Generate sequences using a set of frequencies
#define RAN_SHUFFLE 2 // Generate sequences by shuffling original sequence
#define LONG_SEQ_SIZE 1000000 // Length of the initial pool of randomly generated residues

struct ALPHABET  {
			 char *head;
			 int number;
			 int *size;
			 char **name;
			 char **group;
		  };

#include "lib.h"
#include "math.c"
#include "fasta_file.c"
#include "fsegments.h"

using namespace std;

struct FASTA_FILE f_file;

int window = 0; // Scanning window. Not used if window = 0 

int random_seqs = 0; // Number of random sequences 
int estimate_ran = RAN_OFF; // Type of random sequences used to estimate significance
int max_separation = 5; //Maximum allowed separation for two linked segments
int use_merge = 1; // Merge overlapping or adjacent windows when scanning window is used
int fasta_output = 0; // Generate SEG-like FASTA format file with masked biased segments
int estimate_linkage = 0; // Estimate linkage distance based on expected number of occurrences

char special_aa[4]; // Nucleotides with 'special' properties as specified by the user
                     // !!! 'special_aa' is a '\0'-terminated string  
double P_special; // Probability of observing one 'special' residue (computed from 'aa_freq' using 'special_aa')
double min_p_val = 0.05; // Min. p-value to print a fragment
double masking_p_val = 0.05; // Min. p-value to mask segments in FASTA file

/* Intervals used to generate random sequences */
double interval[6];

/* Average nucleotide frequencies */
double aa_freq[5];

double special_rand_p[MAX_FRAGMENT+1]; // P-value for fragments with high propensity estimated from random seqs

double *random_special_pos; // distribution of the number of positions with high propensity in random sequences

std::vector<FRAGMENT> special,special_merged; // all runs of 'special' amino acids
// Same for random sequences
std::vector<FRAGMENT> r_special;

// Sequences recoded into '1' (special) and '0' (non-special) positions
char *random_binary_seq,
     *binary_seq;
// Sequence with biased segments masked by 'X's
char *masked_seq; 

char nt[] = "ATGC";

// Initial pool of randomly generated residues 
char long_seq[LONG_SEQ_SIZE+1];

/* Allowed command-line arguments */
char valid_args[] = "WESRPNFLATGCMKB";

char Help[] =
"\n---- Copyright (c) 2004-2005 by Igor B. Kuznetsov. All Rights Reserved.----\n"
"\nFind compositionally biased segments in DNA sequence.\n"
"\nNBIAS <FASTA sequence> <output file> [optional arguments]\n"
"\nOptional arguments:\n"
"\n-W:# - scanning window size (OFF by default)"
"\n-E:# - fragments closer than this number of positions will be merged (5 by default)"
"\n-S:# - shuffle the original sequence # times to generate random sequences"
"\n-B:# - use a.a. frequencies to generate # random sequences"
"\n-F   - output SEG-like masked sequence in FASTA format (segments with Prob < min. p-value are masked by N's)"
"\n-L   - output SEG-like masked sequence in FASTA format and mask segments using lower-case letters"
"\n-N   - do not merge overlapping/adjacent scanning windows"
"\n-K   - estimate linkage distance from expected number of occurrences"
"\n-R:string - special residues (default: A)"
"\n-P:# - min. p-value for merging adjacent scanning windows (0.05 by default)"
"\n-M:# - min. p-value for masking FASTA output (0.05 by default)"
"\n-A:# - frequency of A"
"\n-T:# - frequency of T"
"\n-G:# - frequency of G"
"\n-C:# - frequency of C"
"\nNOTE: by default, frequencies are estimated using the input sequence\n";

FILE *out,*fasta_out;

#include "fsegments_dna.cpp"

void InitAll(int argc,char **argv)
{
int i,j;
double sum;
char _str_[1000];
char *freq_tag;
FILE *in;

if(argc < 3) {puts(Help);exit(1);}

if((in=fopen(argv[1],"r")) == NULL) {puts("\nCan't open input file\n");exit(1);}
f_file.fil = in;
if(ReadFastaFile(&f_file) == 0) {puts("\nError reading FASTA file\n");exit(1);}
fclose(in);
Conv2Upper(f_file.seq[0],f_file.l[0]);

// Initialize arrays
for(i = 0;i <= MAX_FRAGMENT;i++) special_rand_p[i] = 0.;

// Set default residues frequencies
strcpy(special_aa,"A");
freq_tag = "Default frequencies";
ComputeNTfreq(f_file.seq[0],f_file.l[0],aa_freq); 

for(i = 3;i < argc;i++) 
{
    if(AAExists(toupper(argv[i][1]),valid_args,strlen(valid_args)) == 0) 
    {printf("\nUnrecognized command line argument: '%c'\n",toupper(argv[i][1]));exit(1);}

    if(argv[i][0] == '-' && toupper(argv[i][1]) == 'W') window = atoi(argv[i]+3);
    if(argv[i][0] == '-' && toupper(argv[i][1]) == 'E') max_separation = atoi(argv[i]+3);
    if(argv[i][0] == '-' && toupper(argv[i][1]) == 'S') {random_seqs = atoi(argv[i]+3);estimate_ran = RAN_SHUFFLE;}
    if(argv[i][0] == '-' && toupper(argv[i][1]) == 'B') {random_seqs = atoi(argv[i]+3);estimate_ran = RAN_FREQ;}
	if(argv[i][0] == '-' && toupper(argv[i][1]) == 'P') min_p_val = atof(argv[i]+3);
	if(argv[i][0] == '-' && toupper(argv[i][1]) == 'M') masking_p_val = atof(argv[i]+3);
	if(argv[i][0] == '-' && toupper(argv[i][1]) == 'N') use_merge = 0;
	if(argv[i][0] == '-' && toupper(argv[i][1]) == 'K') estimate_linkage = 1;
	if(argv[i][0] == '-' && toupper(argv[i][1]) == 'F') fasta_output = 1; // Mask with X's
	if(argv[i][0] == '-' && toupper(argv[i][1]) == 'L') fasta_output = 2; // Mask with lower-case letters 

    // List of special residues
    if(argv[i][0] == '-' && toupper(argv[i][1]) == 'R')
    {
        CopyAA(argv[i]+3,special_aa);
    }

	// Nucleotide frequencies
	if(argv[i][0] == '-' && toupper(argv[i][1]) == 'A') 
		{aa_freq[FindLetter('A',4,nt)] = atof(argv[i]+3);freq_tag = "User-supplied frequencies";}
	if(argv[i][0] == '-' && toupper(argv[i][1]) == 'T') 
		{aa_freq[FindLetter('T',4,nt)] = atof(argv[i]+3);freq_tag = "User-supplied frequencies";}
	if(argv[i][0] == '-' && toupper(argv[i][1]) == 'G') 
		{aa_freq[FindLetter('G',4,nt)] = atof(argv[i]+3);freq_tag = "User-supplied frequencies";}
	if(argv[i][0] == '-' && toupper(argv[i][1]) == 'C') 
		{aa_freq[FindLetter('C',4,nt)] = atof(argv[i]+3);freq_tag = "User-supplied frequencies";}
}

for(sum = 0.,i = 0;i < 4;i++) sum += aa_freq[i];
if(sum < 0.999 || sum > 1.001)  {fprintf(stderr,"ERROR: nucleotide frequencies do not add up to 1.0\n");exit(1);}

if(window > 1 && random_seqs > 0) 
	{fprintf(stderr,"Incompatible switches: scanning window+random sequences\n");exit(1);}
if(use_merge == 0 && window <= 1) 
	{fprintf(stderr,"Incompatible switches: no merge+use linkage\n");exit(1);}
if(use_merge == 0 && window > 1 && fasta_output != 0) 
	{fprintf(stderr,"Incompatible switches: no merge+fasta output\n");exit(1);}

if(window > f_file.l[0]) {printf("\nERROR: window is longer than the sequence\n");exit(1);}

random_special_pos = new double[f_file.l[0]+1];
random_binary_seq = new char[f_file.l[0]+1];
binary_seq = new char[f_file.l[0]+1];
masked_seq = new char[f_file.l[0]+1];

if(random_special_pos == NULL || random_binary_seq == NULL || binary_seq == NULL || masked_seq == NULL) 
	{printf("\nAllocation error in InitAll()\n");exit(1);}

for(i = 0;i <= f_file.l[0];i++) 
{
	random_special_pos[i] = 0.;
	// Recode sequences into binary representation
	binary_seq[i] = random_binary_seq[i] = AA2binary(f_file.seq[0][i],special_aa,strlen(special_aa));
}

P_special = ComputeResiduePNT(special_aa,aa_freq);

// Estimate linkage distance using the expected frequency of special residues
if(estimate_linkage == 1) 
{
	max_separation = EstimateLinkageDistance(P_special);
	printf("Estimated linkage distance = %d\n",max_separation);
}
else printf("Linkage distance = %d\n",max_separation);

init_ran1(0);

// File with runs of 'special' residues
strcpy(_str_,argv[2]);
strcat(_str_,".seg");
if((out=fopen(_str_,"w")) == NULL) {puts("\nCan't open output file\n");exit(1);}

if(window <= 1) fprintf(out,"Length=%d W=%d E=%d\n",f_file.l[0],window,max_separation);
else fprintf(out,"Length=%d W=%d E=%d P=%lf\n",f_file.l[0],window,max_separation,min_p_val);
fprintf(out,"Special residues: %s; %s\n",special_aa,freq_tag);

if(estimate_ran == RAN_SHUFFLE) fprintf(out,"Estimates using %d shuffled sequences\n",random_seqs);
if(estimate_ran == RAN_FREQ) fprintf(out,"Estimates using %d generated sequences\n",random_seqs);

if(fasta_output > 0)
{
	strcpy(_str_,argv[2]);
	strcat(_str_,".fas");
	if((fasta_out=fopen(_str_,"w")) == NULL) {fprintf(stderr,"\nCan't open output file %s\n",_str_);exit(1);}
}

}

void WriteSegmentsNT(char *seq,char *name,int l,std::vector<FRAGMENT>& special_list)
{
int i,j,masked_pos = 0;
int h_n; // The total length of positions occupied by special amino acids
double r_p,d_p,run_p1,run_p2,ab_p,b_p;

if(fasta_output > 0)
{
	strcpy(masked_seq,seq);
	fprintf(fasta_out,">%s\n",name);

}

fprintf(out,"\n>%s\n",name);
form_out_l(out,seq,l,80);

fprintf(out,"\n#Segments: %d\n",special_list.size());

for(i = 0;i < special_list.size();i++)
{
    run_p2 = 0.;
    // Cumulative probability of observing fragment of >= size obtained using unconditional scan statistic
    j = ComputeNSuccesses(special_aa,seq,special_list[i].start-1,special_list[i].end-1);
    //run_p2 = ExactRunProb((int)special_list[i].l,l,P_special);
    run_p1 = P_scan_ge_K_in_W_N(j,(int)special_list[i].l,l,P_special);

	if(random_seqs > 0) run_p1 = CumulativeProb(special_list[i].l,MAX_FRAGMENT,special_rand_p);

	// Skip fragments with p-value above threshold
	// The p-value from random sequences has priority (if random seqs are used)
	//if(random_seqs == 0 && run_p1 > min_p_val) continue;
	//if(random_seqs > 0 && run_p2 > min_p_val) continue;

    // fragment number, start, end, length, number of special residues, p-values (exact and simulation), sequence
    fprintf(out,"%d %d %d %d %d %e ",
                 i+1,special_list[i].start,special_list[i].end,special_list[i].l,j,run_p1);
    for(j = special_list[i].start;j <= special_list[i].end;j++) fprintf(out,"%c",seq[j-1]);
    fprintf(out,"\n");

	// Mark positions in biased segments by 'N's
	if(fasta_output == 1 && run_p1 < masking_p_val)
	{
		for(j = special_list[i].start;j <= special_list[i].end;j++) {masked_seq[j-1] = 'N';masked_pos++;}
	}
	// Mark positions in biased segments by lower-case letters
	if(fasta_output == 2 && run_p1 < masking_p_val)
	{
		for(j = special_list[i].start;j <= special_list[i].end;j++) {masked_seq[j-1] = tolower(seq[j-1]);masked_pos++;}
	}

}

if(fasta_output > 0) 
{
	form_out_l(fasta_out,masked_seq,l,80);
	if(fclose(fasta_out) != 0) fprintf(stderr,"Error closing FASTA file\n");

	printf("Masked positions: %d of %d %.2lf\n",masked_pos,l,100.*(double)masked_pos/(double)l); 
}

// Estimates of the statistical significance of the total number of special positions
int flag;// Marks cases when normal approximation to binomial is not valid with '*'

fprintf(out,"\n");
h_n = 0;

// The total number of special residues
h_n = ComputeNSuccesses(special_aa,seq,0,l-1);

if(h_n >= (int)ceil((double)l*P_special)) // Excess of special positions
{
	b_p = 0;
    //b_p = Binomial_P_GreaterThan_K(h_n,l,P_special);
    ab_p = ApproxBinomial_P_GreaterThan_K(h_n,l,P_special,flag);
	fprintf(out,"#Overall: %d %e %e",h_n,b_p,ab_p);
    if(flag == 0) fprintf(out,"* (+)\n");
    else fprintf(out," (+)\n");
}
else // Lack of special residues
{
	b_p = 0;
    //b_p = Binomial_P_LessThan_K(h_n,l,P_special);
    ab_p = ApproxBinomial_P_LessThan_K(h_n,l,P_special,flag);
	fprintf(out,"#Overall: %d %e %e",h_n,b_p,ab_p);
    if(flag == 0) fprintf(out,"* (-)\n");
    else fprintf(out," (-)\n");
}

}

/* This version goes through each position and exchanges residues with randomly selected position */
void ShuffleSequenceRegular(char *seq,int l)
{
int p,i,j;
char c;

for(j = 0;j < l;j++)
{
        p = (int)ran1((double)l);
        if(p == j) {j--;continue;}
        c = seq[j];
        seq[j] = seq[p];
        seq[p] = c;
}

}

/* Creates intervals from a.a. frequencies used to generate random sequences */
void SetIntervalsNT(double *freq,double *inter)
{
int i;

inter[0] = 0.;
for(i = 1;i < 5;i++) inter[i] = inter[i-1] + freq[i-1];
}

/* Generates random sequence with length 'l' and composition given by 'interval' */
void GenerateRandomSequenceNT(char *seq,int l,double *inter)
{
int i,j;
double r;

for(i = 0;i < l;i++)
{
    r = ran1(1.);
    for(j = 1;j < 5;j++) if(r >= inter[j-1] && r < inter[j]) {seq[i] = aa[j-1];break;}     
}

}

/* Generates random sequence 'dest' of length 'd_l' by sampling characters from 'source' */
void GenerateSequence(char *dest,int d_l,char *source,int s_l)
{
int j,p;

for(j = 0;j < d_l;j++)
{
        p = (int)(ran1(1.)*(double)s_l);
        dest[j] = source[p];
}

}

// Random estimation using linkage 
void EstimateRandomPM(char *seq,int l)
{
int i,j,k,longest,n;
double n_h,n_l,n_b;
std::vector<FRAGMENT>::iterator iter;

// Shuffle the original sequence twice before entering the loop
if(estimate_ran == RAN_SHUFFLE) {ShuffleSequenceRegular(seq,l);ShuffleSequenceRegular(seq,l);}
else {SetIntervalsNT(aa_freq,interval);GenerateRandomSequenceNT(long_seq,LONG_SEQ_SIZE,interval);}


n = strlen(special_aa);
for(i = 0;i < random_seqs;i++)
{
    if((i+1)%1000 == 0) printf("N=%d\n",i+1);

    if(estimate_ran == RAN_SHUFFLE) ShuffleSequenceRegular(seq,l);
    else
    {
	    GenerateSequence(seq,l,long_seq,LONG_SEQ_SIZE);
		for(j = 0;j < l;j++) seq[j] = AA2binary(seq[j],special_aa,n);
    }

	r_special.clear();

    FindRuns(seq,l,r_special);
	longest = MergeSegmentsDiscrete(seq,r_special,max_separation);

    special_rand_p[longest]++;
}

for(i = 0;i <= MAX_FRAGMENT;i++)
{
    special_rand_p[i] /= (double)random_seqs;
}

}

// Random estimation using sliding window
void EstimateRandomPW(char *seq,int l,int w,double p)
{
int i,j,k,max_density,n;
double n_h,n_l,n_b;

// Shuffle the original sequence once before entering the loop
if(estimate_ran == RAN_SHUFFLE) {ShuffleSequenceRegular(seq,l);ShuffleSequenceRegular(seq,l);}
else {SetIntervalsNT(aa_freq,interval);GenerateRandomSequenceNT(long_seq,LONG_SEQ_SIZE,interval);}


n = strlen(special_aa);
for(i = 0;i < random_seqs;i++)
{
    if((i+1)%1000 == 0) printf("N=%d\n",i+1);

    if(estimate_ran == RAN_SHUFFLE) ShuffleSequenceRegular(seq,l);
    else
    {
	    GenerateSequence(seq,l,long_seq,LONG_SEQ_SIZE);
		for(j = 0;j < l;j++) seq[j] = AA2binary(seq[j],special_aa,n);
    }

    max_density = SlideWindow(seq,l,w);

    special_rand_p[max_density]++;
}

for(i = 0;i <= w;i++)
{
    special_rand_p[i] /= (double)random_seqs;
}

}


/******************************************************************************************************/
/*********************************************** MAIN() ***********************************************/
/******************************************************************************************************/

int main(int argc,char **argv)
{
int i;
time_t t1,t2;

t1 = time(NULL);

InitAll(argc,argv);

// Use scanning window
if(window > 1)
{
	SlideWindowP(binary_seq,f_file.l[0],window,min_p_val,special);
	// Merge runs
	special_merged = special;
	if(use_merge == 1) MergeSegmentsDiscrete(f_file.seq[0],special_merged,max_separation);

	if(random_seqs > 0) EstimateRandomPW(random_binary_seq,f_file.l[0],window,min_p_val);
}
// Define window by merging adjacent positions
else
{
	// Find all runs of special residues
	FindRuns(binary_seq,f_file.l[0],special);

	// Merge runs
	special_merged = special;
	MergeSegmentsDiscrete(f_file.seq[0],special_merged,max_separation);
	if(random_seqs > 0) EstimateRandomPM(random_binary_seq,f_file.l[0]);
}

WriteSegmentsNT(f_file.seq[0],f_file.name[0],f_file.l[0],special_merged);

if(fclose(out) != 0) {puts("\nerror closing output file\n");return(1);}

t2 = time(NULL);

printf("\nElapsed time: %d seconds\n",t2-t1);

return(0);
}

